/* 
 * PMach Operating System
 * Copyright (c) 1995 Santa Clara University
 * All Rights Reserved.
 */
/* 
 * Mach Operating System
 * Copyright (c) 1991,1990,1989 Carnegie Mellon University
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 * 
 * CARNEGIE MELLON ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS"
 * CONDITION.  CARNEGIE MELLON DISCLAIMS ANY LIABILITY OF ANY KIND FOR
 * ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 * 
 * Carnegie Mellon requests users of this software to return to
 * 
 *  Software Distribution Coordinator  or  Software.Distribution@CS.CMU.EDU
 *  School of Computer Science
 *  Carnegie Mellon University
 *  Pittsburgh PA 15213-3890
 * 
 * any improvements or extensions that they make and grant Carnegie Mellon
 * the rights to redistribute these changes.
 */
/*
 * Copyright 1996 1995 by Open Software Foundation, Inc. 1997 1996 1995 1994 1993 1992 1991  
 *              All Rights Reserved 
 *  
 * Permission to use, copy, modify, and distribute this software and 
 * its documentation for any purpose and without fee is hereby granted, 
 * provided that the above copyright notice appears in all copies and 
 * that both the copyright notice and this permission notice appear in 
 * supporting documentation. 
 *  
 * OSF DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE 
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
 * FOR A PARTICULAR PURPOSE. 
 *  
 * IN NO EVENT SHALL OSF BE LIABLE FOR ANY SPECIAL, INDIRECT, OR 
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM 
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, 
 * NEGLIGENCE, OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION 
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 
 * 
 */
/* 
 * MkLinux_DR2.1_Update1
 */
/*
  Copyright 1990 by Open Software Foundation,
Cambridge, MA.

		All Rights Reserved

  Permission to use, copy, modify, and distribute this software and
its documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appears in all copies and
that both the copyright notice and this permission notice appear in
supporting documentation, and that the name of OSF or Open Software
Foundation not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

  OSF DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE
INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
IN NO EVENT SHALL OSF BE LIABLE FOR ANY SPECIAL, INDIRECT, OR
CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT,
NEGLIGENCE, OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION
WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <debug.h>
#include <lan.h>
#include <ppc/POWERMAC/enet_dma.h>
#include <ppc/POWERMAC/if_am79940.h>

#include	<kern/time_out.h>
#include	<device/device_types.h>
#include	<device/ds_routines.h>
#include	<device/errno.h>
#include	<device/io_req.h>
#include	<device/if_hdr.h>
#include	<device/if_ether.h>
#include	<device/net_status.h>
#include	<device/net_io.h>
#include	<machine/endian.h> /* for ntohl */

#include	<ppc/spl.h>
#include	<chips/busses.h>

int	at3c501probe();
void	at3c501attach();
int	at3c501intr();
int	at3c501init();
int	at3c501output();
int	at3c501ioctl();
int	at3c501reset();
int	at3c501watch();

void mace_init();
void rcv_init();
void xmt_init();
void iodma_base_init();
void printbuf();

static caddr_t at3c501_std[NLAN] = { 0 };
static struct bus_device *at3c501_info[NLAN];
struct	bus_driver	at3c501driver = 
	{(probe_t)at3c501probe, 0,
	 (void (*)(struct bus_device*))at3c501attach, 0,
         at3c501_std, "et", at3c501_info, };

int	watchdog_id;

typedef struct { 
	struct	ifnet	ds_if;		/* generic interface header */
	u_char	ds_addr[6];		/* Ethernet hardware address */

	int	flags;
        int     timer;
	char 	*base;
        u_char   address[ETHER_ADD_SIZE];
	short	mode;
	int	badxmt;
	int	badrcv;
	int	spurious;
	int	rcv;
	int	xmt;
} at3c501_softc_t;

at3c501_softc_t	at3c501_softc[NLAN];

#if DEBUG
#define LANDEBUG_TRACE   1	/* set if we want debugging tracing */
#define LANDEBUG_VERBOSE 2	/* set for packet contents too */
#define LANDEBUG_ENABLED 4	/* (sticky) set when an error occurs */

#if 1
int landebug=0;
#else
int landebug=LANDEBUG_TRACE; /* tracing enabled only on error */
#endif /* 1 */

#define KGDB_PRINTF(A) if ((landebug & LANDEBUG_TRACE) && \
			   (landebug & LANDEBUG_ENABLED)) { kgdb_printf A; }
#else
#define KGDB_PRINTF(A)
#endif

/*
 * at3c501probe:
 *
 *	This function "probes" or checks for the 3c501 board on the bus to see
 *	if it is there.  As far as I can tell, the best break between this
 *	routine and the attach code is to simply determine whether the board
 *	is configured in properly.  Currently my approach to this is to write
 *	and read a string from the Packet Buffer on the board being probed.
 *	If the string comes back properly then we assume the board is there.
 *	The config code expects to see a successful return from the probe
 *	routine before 	attach will be called.
 *
 * input	: address device is mapped to, and unit # being checked
 * output	: a '1' is returned if the board exists, and a 0 otherwise
 *
 */
at3c501probe(port, dev)
struct bus_device	*dev;
{
	caddr_t		base = (caddr_t)dev->address;
	int		unit = dev->unit;
	char		inbuf[50];
	char		*str = "3c501 ethernet board %d out of range\n";
	int 		strsize = strlen(str);

	KGDB_PRINTF(("It is at3c501 probe\n"));
	KGDB_PRINTF(("ereg base address is %08x\n", base));
		
	ereg = (struct am79c940 *)base;
		
	if ((unit < 0) || (unit >= NLAN)) {
		printf(str, unit);
		return(0);
	}

	/* reset */
	/* outb(IE_CSR(base), IE_RESET); */
	
	ereg->biucc = BIUCC_SWRST;
	

	/* write a string to the packet buffer */
	/* outb(IE_CSR(base), IE_RIDE | IE_SYSBFR); */
	/* outw(IE_GP(base), 0); */
	/* loutb(IE_BFR(base), str, strsize); */

	/* read it back */
	/* outb(IE_CSR(base), IE_RIDE | IE_SYSBFR); */
	/* outw(IE_GP(base), 0); */
	/* linb(IE_BFR(base), inbuf, strsize); */
	/* compare them */

#if 0 /* TODO NMGS - clean up or remove */

	if (strncmp(str, inbuf, strsize))
	{
		return(0);
	}
#endif 
	
	if ((ereg->xmtfc != 0x01) && (ereg->rcvfc != 0x01))
		return(0);
		
	at3c501_softc[unit].base = base;

	return(1);
}

/*
 * at3c501attach:
 *
 *	This function attaches a 3C501 board to the "system".  The rest of
 *	runtime structures are initialized here (this routine is called after
 *	a successful probe of the board).  Once the ethernet address is read
 *	and stored, the board's ifnet structure is attached and readied.
 *
 * input	: bus_device structure setup in autoconfig
 * output	: board structs and ifnet is setup
 *
 */
void at3c501attach(dev)
struct bus_device	*dev;
{
	at3c501_softc_t	*sp;
	struct	ifnet	*ifp;
	u_char	unit;

	caddr_t		base;

	KGDB_PRINTF(("It is at3c501attach\n"));
	KGDB_PRINTF(("IO_DMA_BASE is %08x\n", IO_DMA_BASE));
	KGDB_PRINTF(("IO_DMA_BASE_V is %08x\n", IO_DMA_BASE_V));
	KGDB_PRINTF(("dma_buffer_r is %08x\n", dma_buffer_r));
	KGDB_PRINTF(("dma_buffer_v is %08x\n", dma_buffer_v));

	/* take_dev_irq(dev);  old */
	
	/*  initialize interrupt handler */
	/* set_at3c501_ih(); */
	
	unit = (u_char)dev->unit;	
	KGDB_PRINTF((", port = %x, spl = %d, pic = %d. ",
		dev->address, dev->sysdep, dev->sysdep1));

	sp = &at3c501_softc[unit];
	base = sp->base;
	
	/* ereg = (struct am79c940*)base */

	if (base != (caddr_t)dev->address) {
		printf("3C501 board %d attach address error\n", unit);
		return;
	}
	sp->timer = -1;
	sp->flags = 0;
	sp->mode = 0;
	/* outb(IE_CSR(sp->base), IE_RESET); */
	/* at3c501geteh(base, sp->ds_addr); */
	/* at3c501geteh(base, sp->address); */
	/* at3c501seteh(base, sp->address); */

	at3c501geteh(base, sp->ds_addr);
	at3c501geteh(base, sp->address);
	
	printf("ethernet id [%02x:%02x:%02x:%02x:%02x:%02x] ",
		sp->address[0], sp->address[1], sp->address[2], 
		sp->address[3], sp->address[4], sp->address[5]);
	ifp = &(sp->ds_if);
	ifp->if_unit = unit;
	ifp->if_mtu = ETHERMTU;
	ifp->if_flags = IFF_BROADCAST;

	ifp->if_header_size = sizeof(struct ether_header);
	ifp->if_header_format = HDR_ETHERNET;
	ifp->if_address_size = 6;
	ifp->if_address = (char *)&sp->address[0];
	if_init_queues(ifp);
	ifp->if_snd.ifq_maxlen = 8;/* Added NMGS */
}

/*
 * at3c501watch():
 *
 */
at3c501watch(b_ptr)

caddr_t	b_ptr;

{
	int	x,
		y,
		opri,
		unit;
	at3c501_softc_t *is;

	KGDB_PRINTF(("It is at3c501watch()\n"));
	
	unit = *b_ptr;

	timeout((timeout_fcn_t)at3c501watch,b_ptr,20*hz);

	is = &at3c501_softc[unit];
	printf("\nxmt/bad	rcv/bad	spurious\n");
	printf("%d/%d		%d/%d	%d\n", is->xmt, is->badxmt, \
		is->rcv, is->badrcv, is->spurious);
	is->rcv=is->badrcv=is->xmt=is->badxmt=is->spurious=0;
}

/*
 * at3c501geteh:
 *
 *	This function gets the ethernet address (array of 6 unsigned
 *	bytes) from the 3c501 board prom. 
 *
 */

at3c501geteh(base, ep)
caddr_t	base;
char *ep;
{
	int	i;
	unsigned char ep_temp;

	KGDB_PRINTF(("It is at3c501geteh()\n"));
	
	ereg->iac = IAC_PHYADDR; delay(100);
	
	KGDB_PRINTF(("\nethernet id is "));
	
	for (i = 0; i < ETHER_ADD_SIZE; i++) {
		/* outw(IE_GP(base), i); */
		/* *ep++ = inb(IE_SAPROM(base)); */
		
		ep_temp = ereg->padr;
		KGDB_PRINTF(("%02x ", ep_temp));
		*ep++ = ep_temp;
	}
	KGDB_PRINTF(("\n\n"));
}

/*
 * at3c501seteh:
 *
 *	This function sets the ethernet address (array of 6 unsigned
 *	bytes) on the 3c501 board. 
 *
 */

at3c501seteh(base, ep)
caddr_t	base;
char *ep;
{
	int	i;

	for (i = 0; i < ETHER_ADD_SIZE; i++) {
		/* outb(EDLC_ADDR(base) + i, *ep++); */
	}
}

void at3c501start(int);	/* forward */

at3c501output(dev, ior)
	dev_t		dev;
	io_req_t	ior;
{
	register int	unit = minor(dev);
	int  res;
	int  oldmsr;
	
KGDB_PRINTF(("it is at3c501output()\n"));
	if (unit < 0 || unit >= NLAN ||
		at3c501_softc[unit].base == 0)
	    return (ENXIO);
    
	res = net_write(&at3c501_softc[unit].ds_if, at3c501start, ior);
	
	return res;
}

at3c501setinput(dev, receive_port, priority, filter, filter_count, device)
	dev_t		dev;
	mach_port_t	receive_port;
	int		priority;
	filter_t	filter[];
	u_int		filter_count;
	device_t	device;
{
	register int unit = minor(dev);

	KGDB_PRINTF(("It is at3c501setinput()\n"));

	if (unit < 0 || unit >= NLAN ||
		at3c501_softc[unit].base == 0)
	    return (ENXIO);

	return (net_set_filter(&at3c501_softc[unit].ds_if,
			(ipc_port_t)receive_port, priority,
			filter, filter_count, device));
}


/*
 * at3c501reset:
 *
 *	This routine is in part an entry point for the "if" code.  Since most 
 *	of the actual initialization has already (we hope already) been done
 *	by calling at3c501attach().
 *
 * input	: unit number or board number to reset
 * output	: board is reset
 *
 */
at3c501reset(unit)
int	unit;
{
	KGDB_PRINTF(("It is at3c501reset()\n"));
	
	at3c501_softc[unit].ds_if.if_flags &= ~IFF_RUNNING;
	return(at3c501init(unit));
}



/*
 * at3c501init:
 *
 *	Another routine that interfaces the "if" layer to this driver.  
 *	Simply resets the structures that are used by "upper layers".  
 *	As well as calling at3c501hwrst that does reset the at3c501 board.
 *
 * input	: board number
 * output	: structures (if structs) and board are reset
 *
 */	
at3c501init(unit)
int	unit;
{
	struct	ifnet	*ifp;
	int		stat;
	spl_t		oldpri;

	KGDB_PRINTF(("It is at3c501init()\n"));

	ifp = &(at3c501_softc[unit].ds_if);

	oldpri = splnet();
	if ((stat = at3c501hwrst(unit)) == TRUE) {
		at3c501_softc[unit].ds_if.if_flags |= IFF_RUNNING;
		at3c501_softc[unit].flags |= DSF_RUNNING;
		at3c501start(unit);
	}
	else
		printf("3C501 trouble resetting board %d\n", unit);
	at3c501_softc[unit].timer = 5;
	splx(oldpri);
	return(stat);

}

/*ARGSUSED*/
at3c501open(dev, flag)
	dev_t	dev;
	int	flag;
{
	register int	unit = minor(dev);
	
	KGDB_PRINTF(("It is at3c501open()\n"));
	
	if (unit < 0 || unit >= NAT3C501 ||
		at3c501_softc[unit].base == 0)
	    return (ENXIO);

	/* interrupt_init();   ????????????? */

	at3c501_softc[unit].ds_if.if_flags |= IFF_UP;
	at3c501init(unit);
	return(0);
}

/*
 * at3c501start:
 *
 *	This is yet another interface routine that simply tries to output a
 *	in an mbuf after a reset.
 *
 * input	: board number
 * output	: stuff sent to board if any there
 *
 */
void at3c501start(int unit)
{

	io_req_t	m;

	struct	ifnet	*ifp;

 	KGDB_PRINTF(("It is at3c501start()\n")); 

	ifp = &(at3c501_softc[unit].ds_if);
	for(;;) {
		IF_DEQUEUE(&ifp->if_snd, m);

		if (m != 0)
			at3c501xmt(unit, m);
		else
			return;
	}
}

/*ARGSUSED*/
at3c501getstat(dev, flavor, status, count)
	dev_t		dev;
	int		flavor;
	dev_status_t	status;		/* pointer to OUT array */
	u_int		*count;		/* out */
{
	register int	unit = minor(dev);

	KGDB_PRINTF(("It is at3c501getstat()\n"));

	if (unit < 0 || unit >= NAT3C501 ||
		at3c501_softc[unit].base == 0)
	    return (ENXIO);

	return (net_getstat(&at3c501_softc[unit].ds_if,
			    flavor,
			    status,
			    count));
}

at3c501setstat(dev, flavor, status, count)
	dev_t		dev;
	int		flavor;
	dev_status_t	status;
	u_int		count;
{
	register int	unit = minor(dev);
	register at3c501_softc_t *sp;

	KGDB_PRINTF(("It is at3c501setstat()\n"));

	if (unit < 0 || unit >= NAT3C501 ||
		at3c501_softc[unit].base == 0)
	    return (ENXIO);

	sp = &at3c501_softc[unit];

	switch (flavor) {
	    case NET_STATUS:
	    {
		/*
		 * All we can change are flags, and not many of those.
		 */
		register struct net_status *ns = (struct net_status *)status;
		int	mode = 0;

		if (count < NET_STATUS_COUNT)
		    return (D_INVALID_SIZE);

		if (ns->flags & IFF_ALLMULTI)
		    mode |= MOD_ENAL;
		if (ns->flags & IFF_PROMISC)
		    mode |= MOD_PROM;

		/*
		 * Force a compilete reset if the receive mode changes
		 * so that these take effect immediately.
		 */
		if (sp->mode != mode) {
		    sp->mode = mode;
		    if (sp->flags & DSF_RUNNING) {
			sp->flags &= ~(DSF_LOCK | DSF_RUNNING);
			at3c501init(unit);
		    }
		}
		break;
	    }
	    case NET_ADDRESS:
	    {
		register union ether_cvt {
		    char	addr[6];
		    int		lwd[2];
		} *ec = (union ether_cvt *)status;

		if (count < sizeof(*ec)/sizeof(int))
		    return (D_INVALID_SIZE);

		ec->lwd[0] = ntohl(ec->lwd[0]);
		ec->lwd[1] = ntohl(ec->lwd[1]);
   		at3c501seteh(sp->base, ec->addr);
		break;
	    }

	    default:
		return (D_INVALID_OPERATION);
	}
	return (D_SUCCESS);
}


/*
 * at3c501hwrst:
 *
 *	This routine resets the at3c501 board that corresponds to the 
 *	board number passed in.
 *
 * input	: board number to do a hardware reset
 * output	: board is reset
 *
 */
#define XMT_STAT (EDLC_16|EDLC_JAM|EDLC_UNDER|EDLC_IDLE)
#define RCV_STAT (EDLC_STALE|EDLC_ANY|EDLC_SHORT|EDLC_DRIBBLE|EDLC_OVER|EDLC_FCS)
int 
at3c501hwrst(unit)
int unit;
{
	u_char	stat;
	caddr_t	base = at3c501_softc[unit].base;

	KGDB_PRINTF(("It is at3c501hwrst()\n"));

#if 0 /* TODO NMGS clean up or remove */
	outb(IE_CSR(base), IE_RESET);
	outb(IE_CSR(base), 0);
	at3c501seteh(base, at3c501_softc[unit].address);
	if ((stat = inb(IE_CSR(base))) != IE_RESET) {
		printf("at3c501reset: can't reset CSR: %x\n", stat);
		return(FALSE);
	}
	if ((stat = inb(EDLC_XMT(base))) & XMT_STAT) {
		printf("at3c501reset: can't reset XMT: %x\n", stat);
		return(FALSE);
	}
	if (((stat = inb(EDLC_RCV(base))) & RCV_STAT) != EDLC_STALE) {
		printf("at3c501reset: can't reset RCV: %x\n", stat);
		return(FALSE);
	}
	if (at3c501config(unit) == FALSE) {
		printf("at3c501hwrst(): failed to config\n");
		return(FALSE);
	}
	outb(IE_RP(base), 0);
	outb(IE_CSR(base), IE_RIDE|IE_RCVEDLC);
	return(TRUE);
#endif /* 0 */
	
	iodma_base_init();

	mace_init();
	xmt_init();	
	rcv_init();
	
	return(TRUE);
}

/*
 * at3c501intr:
 *
 *	This function is the interrupt handler for the at3c501 ethernet
 *	board.  This routine will be called whenever either a packet
 *	is received, or a packet has successfully been transfered and
 *	the unit is ready to transmit another packet.
 *
 * input	: board number that interrupted
 * output	: either a packet is received, or a packet is transfered
 *
 */
at3c501intr(unit)
int unit;
{
	if (edma->rcvcs & OVERRUN) {
#if DEBUG
		landebug|=LANDEBUG_ENABLED;
#endif /* DEBUG */
		KGDB_PRINTF(("at3c501intr() detects overrun\n"));
	} else {
		edma->rcvcs = DMARUN | IF; /* ???????????????? */
	}
KGDB_PRINTF(("It is at3c501intr()\n"));

 	at3c501start(unit);
	at3c501rcv(unit);
	
	if (edma->rcvcs & OVERRUN) {
		KGDB_PRINTF(("at3c501intr() detects overrun (2)\n"));
	} else {
		edma->rcvcs = DMARUN | IE; /* ???????????????? */
	}

	return(0);
}


/*
 * at3c501rcv:
 *
 *	This routine is called by the interrupt handler to initiate a
 *	packet transfer from the board to the "if" layer above this
 *	driver.  This routine checks if a buffer has been successfully
 *	received by the at3c501.  If so, the routine at3c501read is called
 *	to do the actual transfer of the board data (including the
 *	ethernet header) into a packet (consisting of an mbuf chain).
 *
 * input	: number of the board to check
 * output	: if a packet is available, it is "sent up"
 *
 */
 
/*  new rcv function */

at3c501rcv(unit)
	int	unit;
{
	int stat;
	caddr_t base;
	ipc_kmsg_t	new_kmsg;
	struct ether_header *ehp;
	struct packet_header *pkt;
	u_short	len;
	int truelen;
	register struct ifnet *ifp;
	struct ether_header header;
	unsigned char    error_flag = 0;
	register at3c501_softc_t	*is;
	register struct	ifqueue		*inq;
	spl_t	opri;
	struct	ether_header eh;
	
	unsigned char *rcvtp, *rcvtp_v, *status, *tpadj, *packet_header, rcvcs;
	
	KGDB_PRINTF(("It is at3c501rcv()\n")); 
	
	is = &at3c501_softc[unit];
	ifp = &is->ds_if;
	base = at3c501_softc[unit].base;
	is->rcv++;
	
	if ((edma->rcvhp == edma->rcvtp) && ((edma->rcvcs & OVERRUN)==0)) {
#if DEBUG
		landebug|=LANDEBUG_ENABLED;
#endif /* DEBUG */
		KGDB_PRINTF(("No packet received, no overrun\n"));

		at3c501init(unit);

#if DEBUG
		landebug&=~LANDEBUG_ENABLED;
#endif /* DEBUG */
		edma->rcvcs = DMARUN | IE;
		return;
	}
	
	do {
		
		KGDB_PRINTF(("before processing packet\n"));
 		KGDB_PRINTF(("h=%02x ",(volatile unsigned char)edma->rcvhp));
 		KGDB_PRINTF(("t=%02x\n ",(volatile unsigned char)edma->rcvtp));
		
		rcvtp   = (unsigned char*)(DMA_RCV_BASE  +(edma->rcvtp << 8));
		rcvtp_v = (unsigned char*)(DMA_RCV_BASE_V +(edma->rcvtp << 8));
		
		status = rcvtp_v;
		
		len = *status++ << 8;
		len = len + *status;
		
		if (len & 0xf000) {
#if DEBUG
			landebug|=LANDEBUG_ENABLED;
#endif /* DEBUG */
			KGDB_PRINTF(("Error in rcv status register:%x\n",
				     len));
			at3c501init(unit);
#if DEBUG
			landebug&=~LANDEBUG_ENABLED;
#endif /* DEBUG */
			goto next;
		}
		len = len & 0x0fff;
 		KGDB_PRINTF(("received packet's length is %04x\n", len)); 
#if 0
 		if (len > 1518) {
			printbuf(rcvtp_v, len + 8 );
		}
#endif					
		packet_header = rcvtp_v + 8;
		
		rcvtp = rcvtp + 8 + len; /* status + frame */
		rcvtp = (char*)(((unsigned)rcvtp + 0xff) & ~0xff);
		
		if (rcvtp >= (unsigned char*)(DMA_RCV_BASE + DMA_RCV_SIZE)) {
			KGDB_PRINTF(("original rcvtp = %08x \n", rcvtp));

			/* we've overrun circular buffer, wrap around */

			rcvtp -= DMA_RCV_SIZE;
			assert((rcvtp >= (unsigned char*)DMA_RCV_BASE) &&
			       (rcvtp <	(unsigned char*)
					 (DMA_RCV_BASE + DMA_RCV_SIZE)));
			KGDB_PRINTF(("rounding rcvtp = %08x \n", rcvtp));
		}
		
		KGDB_PRINTF(("after processing packet\n"));
		KGDB_PRINTF(("edma->rcvhp is %02x \n", edma->rcvhp));
		KGDB_PRINTF(("edma->rcvtp is %02x \n", edma->rcvtp));
	    	KGDB_PRINTF(("&eh = %x , packet_header = %x \n", &eh, packet_header));
		
		if (error_flag) {
#if DEBUG
			landebug|=LANDEBUG_ENABLED;
#endif /* DEBUG */
			KGDB_PRINTF(("error flag set, ignoring packet\n"));
			edma->rcvtp = ((unsigned long)rcvtp >> 8) & 0xff;
			error_flag = 0;
			goto next;
		}
	    	truelen = len - sizeof(struct ether_header) - 4;

		if (truelen < 0) { 
#if DEBUG
			landebug|=LANDEBUG_ENABLED;
#endif /* DEBUG */
			is->ds_if.if_rcvdrops++;
			KGDB_PRINTF(("Error: received packet is too short (len=%d, probably cause infinite loop\n",len));
			edma->rcvtp = ((unsigned long)rcvtp >> 8) & 0xff;
			at3c501init(unit);
#if DEBUG
			landebug&=~LANDEBUG_ENABLED;
#endif /* DEBUG */
			goto next;
		}
		
		/* packet header is within first 256 bytes, no risk of wrap */

		bcopy(packet_header, (char*)&eh,  sizeof(struct ether_header));
		
		new_kmsg = net_kmsg_get();
		if (new_kmsg == IKM_NULL) {
#if DEBUG
			landebug|=LANDEBUG_ENABLED;
#endif /* DEBUG */
			KGDB_PRINTF(("new_kmsg is IKM_NULL....\n"));
			is->ds_if.if_rcvdrops++;
			
			edma->rcvtp = ((unsigned long)rcvtp >> 8) & 0xff;
			/* printf("after processing packet\n"); */
			/* printf("edma->rcvhp is %02x \n", edma->rcvhp);	 */
			/* printf("edma->rcvtp is %02x \n", edma->rcvtp); */
			
			/* edma->rcvcs = DMARUN | IE; //??????????????			 */
			goto next;
		}
		ehp = (struct ether_header *)(&net_kmsg(new_kmsg)->header[0]);		
		pkt = (struct packet_header *)(&net_kmsg(new_kmsg)->packet[0]);
		
		*ehp = eh;

		KGDB_PRINTF(("copying received packet to %x\n", (char *)(pkt+1))); 
		
		/* Check if new backpointer is before current packet,
		 * ie. the backpointer has wrapped around the circular buffer.
		 * If so, and if new backpointer is after start of circular
		 * buffer, then a packet is split over the wraparound
		 */
		if ((rcvtp < packet_header) &&
		    (rcvtp > (unsigned char*)DMA_RCV_BASE)) {

			/* We've got a packet that is wrapped around the
			 * end of the circular buffer, so it must be
			 * copied in two parts
			 */
			char *ptr;
			int  mid_len;

			ptr = packet_header + sizeof(struct ether_header);

			KGDB_PRINTF(("WRAPAROUND on buffer size %d"
				     "(with header+crc,size=%d)\n",
				     truelen,len));
			mid_len = DMA_RCV_SIZE - ((unsigned int)ptr &
						  (DMA_RCV_SIZE-1));
			/* This gets even more complicated - the last
			 * 4 bytes aren't copied as they're the CRC,
			 * and the wraparound may occur here. Deal with
			 * this too!
			 */
			assert ((mid_len > 0) && (mid_len < (truelen+4)));

			if (mid_len >= truelen) {
				/* The wraparound occurs on the 4 byte CRC
				 * only copy stuff in the actual packet
				 */
				bcopy(ptr, (char *)(pkt + 1), truelen);
			} else {
				/* The wraparound occurs mid-packet, copy
				 * the two halves
				 */

				/* make sure we don't copy too much */
				assert(mid_len < truelen);

				bcopy(ptr, (char *)(pkt + 1), mid_len);
				bcopy((char*)DMA_RCV_BASE,
				      (char *)(pkt + 1) + mid_len,
				      (truelen - mid_len));
			}
		} else {
			/* Normal case - buffer is contiguous */

			bcopy(packet_header + sizeof(struct ether_header),
			      (char *)(pkt + 1),
			      truelen);
		}

		pkt->type = ehp->ether_type;
		pkt->length = len - 4 - sizeof(struct ether_header) + sizeof(struct packet_header);
		
		edma->rcvtp = ((unsigned long)rcvtp >> 8) & 0xff;
		
		KGDB_PRINTF(("after processing packet\n")); 
		KGDB_PRINTF(("edma->rcvhp is %02x \n", edma->rcvhp));
		KGDB_PRINTF(("edma->rcvtp is %02x \n", edma->rcvtp));
		/* 		    if (!ethernet_priority(new_kmsg)) printf("B"); */
		net_packet(ifp, new_kmsg, pkt->length,
			   ethernet_priority(new_kmsg), (io_req_t)0);
next:
		
	} while (edma->rcvhp != edma->rcvtp);

	/* edma->rcvcs = DMARUN | IE;  //??????????????? */
	
}		

		
		
	


/*
 * at3c501xmt:
 *
 *	This routine fills in the appropriate registers and memory
 *	locations on the 3C501 board and starts the board off on
 *	the transmit.
 *
 * input	: board number of interest, and a pointer to the mbuf
 * output	: board memory and registers are set for xfer and attention
 *
 */


/* new xmt function */

at3c501xmt(unit, m)
int	unit;
io_req_t	m;
{
	int			i;
	at3c501_softc_t		*is = &at3c501_softc[unit];
	caddr_t			base = is->base;
	u_short			count = 0;
	u_short			bytes_in_msg;

	unsigned char   cs, xmtrc, xmtfs;
	unsigned char   *xmt_base, *ch_ptr;
	vm_offset_t	phys_xmt_base;
	
KGDB_PRINTF(("It is at3c501xmt()++++++++++++++++++++++++++++++++++++++++++\n")); 

	is->xmt++;	
	count = m->io_count;
	
#define max(a, b) 	(((a) > (b)) ? (a) : (b))

	bytes_in_msg = max(count, ETHERMIN + sizeof(struct ether_header));
	
	/* printf("data buffer in m\n"); */
	/* printbuf(m->io_data, count); */
	
	for ( ; ; ) {
		cs = edma->xmtcs;
		if (cs & SET_0) {
			xmt_base = (unsigned char*)DMA_XMT_BASE_0_V;
			phys_xmt_base = DMA_XMT_BASE_0;
			break;
		} /*else if (cs & SET_1) {
			xmt_base = (unsigned char*)DMA_XMT_BASE_1_V;
			phys_xmt_base = DMA_XMT_BASE_1;
			break;
		} */
	}
	
	bcopy(m->io_data, xmt_base, count);
	
	if ( count < bytes_in_msg ) {
		ch_ptr = xmt_base + count;
		for ( i = 0; i < bytes_in_msg - count; i++ ) 
			*ch_ptr++ = 0x00;
		count = bytes_in_msg;
	}	

	flush_cache(phys_xmt_base, count);
	/* Make sure the DMA sees the real data, which otherwise might be
	   hiding out in the processor's cache.  */

	/* printf("count is %d\n", count); */
  	/* printf("DMA_XMT_BASE_0 is %08x \n", DMA_XMT_BASE_0); */
  	/* printf("DMA_XMT_BASE_1 is %08x \n", DMA_XMT_BASE_1); */
  	/* printf("DMA_XMT_BASE_0_V is %08x \n", DMA_XMT_BASE_0_V); */
  	/* printf("DMA_XMT_BASE_1_V is %08x \n", DMA_XMT_BASE_1_V); */

 	KGDB_PRINTF(("%02x bytes being transmitted\n", count)); 
#if DEBUG
      	 printbuf(xmt_base, count);
#endif /* DEBUG */

#define 	DEBUG_1 	0
#if DEBUG_1
  KGDB_PRINTF(("beore xmt\n"));
  KGDB_PRINTF(("xmt dma count_0 is %02x%02x \n", edma->xmtbch_0 & 0x07, edma->xmtbcl_0));
  KGDB_PRINTF(("xmt dma count_1 is %02x%02x \n", edma->xmtbch_1 & 0x07, edma->xmtbcl_1));  
  KGDB_PRINTF(("edma->xmtcs is %02x \n", edma->xmtcs));
  KGDB_PRINTF(("ereg->pr    is %02x \n", ereg->pr   ));
  KGDB_PRINTF(("ereg->xmtrc is %02x \n", ereg->xmtrc));  
  KGDB_PRINTF(("ereg->xmtfs is %02x \n", ereg->xmtfs));  
  KGDB_PRINTF(("ereg->maccc is %02x \n", ereg->maccc));
#endif  


  xmtrc = ereg->xmtrc;			   
  xmtfs = ereg->xmtfs;


   edma->xmtcs &= 0xfd;  /* look out when using dual buffer! */
   
	if ((int)xmt_base == DMA_XMT_BASE_0_V) {
		/* printf("xmt_base is xmt_0\n"); */
		/* *(volatile unsigned short *) &edma->xmtbch_0 = count; */
		edma->xmtbch_0 = count >> 8;
		edma->xmtbcl_0 = count;
	} else {
		/* printf("xmt_base is xmt_1\n"); */
		/* *(volatile unsigned short *) &edma->xmtbch_1 = count; */
		edma->xmtbch_1 = count >> 8;
		edma->xmtbcl_1 = count;
	}
   edma->xmtcs |= 0x02;
	
/* 	delay(100); */
	
#if DEBUG_1
  KGDB_PRINTF(("after xmt\n"));
  KGDB_PRINTF(("xmt dma count_0 is %02x%02x \n", edma->xmtbch_0 & 0x07, edma->xmtbcl_0));
  KGDB_PRINTF(("xmt dma count_1 is %02x%02x \n", edma->xmtbch_1 & 0x07, edma->xmtbcl_1));  
  KGDB_PRINTF(("edma->xmtcs is %02x \n", edma->xmtcs));
  KGDB_PRINTF(("ereg->pr    is %02x \n", ereg->pr   ));
#endif
  do {
  xmtrc = ereg->xmtrc;			   
  xmtfs = ereg->xmtfs;
  } while ((xmtfs & 0x80) == 0);

#if DEBUG_1
  KGDB_PRINTF(("ereg->xmtrc is %02x \n", xmtrc));  
  KGDB_PRINTF(("ereg->xmtfs is %02x \n", xmtfs));  
  KGDB_PRINTF(("ereg->maccc is %02x \n", ereg->maccc));
#endif  
  if ((xmtfs & 0x80) == 0) {
    KGDB_PRINTF(("I don't know how to process this error ! xmtfs still not vaild after delay(100)!\n"));
  } else {
	if ((xmtfs & 0x62) != 0) {
		printf("Error in XMTFS of 79c940:%x\n", xmtfs);
		at3c501init(unit);
	}
		
  }
	iodone(m);

}




/*
 * At3c501config:
 *
 *	This routine does a standard config of the at3c501 board.
 *
 */
at3c501config(unit)
int	unit;
{
	caddr_t	base = at3c501_softc[unit].base;
	u_char	stat;

	/* Enable DMA & Interrupts */
#if 0

	outb(IE_CSR(base), IE_RIDE|IE_SYSBFR);

	No Transmit Interrupts

	outb(EDLC_XMT(base), 0);
	inb(EDLC_XMT(base));

	Setup Receive Interrupts

	outb(EDLC_RCV(base), EDLC_BROAD|EDLC_SHORT|EDLC_GOOD|EDLC_DRIBBLE|EDLC_OVER);
	inb(EDLC_RCV(base));

	outb(IE_CSR(base), IE_RIDE|IE_SYSBFR);
	outb(IE_RP(base), 0);
	outb(IE_CSR(base), IE_RIDE|IE_RCVEDLC);
	return(TRUE);
#endif /* 0 */
}

/*
 * at3c501intoff:
 *
 *	This function turns interrupts off for the at3c501 board indicated.
 *
 */
at3c501intoff(unit)
int unit;
{
	caddr_t base = at3c501_softc[unit].base;
	/* outb(IE_CSR(base), 0); */
	
	KGDB_PRINTF(("It is at3c501intoff()\n"));	
	ereg->imr = 0xff;
}




int inb(char *s)
{
}

int outb(char *s, int i)
{
}

int outw(char *s, int i)
{
}

int inw(char *s)
{
}

int linb(char *s, struct ether_header *eh, unsigned int i)
{
}

int loutb(char *s1, char *s2, int i)
{
}



void mace_init()
{
  ereg->biucc = BIUCC_SWRST; delay(100);
  
  ereg->imr   = 0xff; /* ???? ~(IMR_MXMTINT);*/
  
  ereg->phycc  = PHYCC_ASEL;

  /* ereg->iac   = PHYADDR; delay(100);   */
  /* GETADDR(enet_phyaddr) */
  /* PRNADDR(enet_phyaddr) */

  ereg->utr    = UTR_NOLOOP;
  
  ereg->maccc  = MACCC_ENXMT | MACCC_ENRCV;
  
 /* ereg->rcvfc  = RCVFC_M_R | RCVFC_ASTRPRCV; */
}


void iodma_base_init()
{

  iodma_base->addr_1 = (IO_DMA_BASE >> 24) & 0xff; 
  iodma_base->addr_2 = (IO_DMA_BASE >> 16) & 0xff; 
  iodma_base->addr_3 = (IO_DMA_BASE >>  8) & 0xff; 
  iodma_base->addr_4 = (IO_DMA_BASE      ) & 0xff; 

#define DEBUG_DMABASE 1
#if DEBUG_DMABASE
  KGDB_PRINTF(("dma_buffer_r is %08x \n", dma_buffer_r));
  KGDB_PRINTF(("IO_DMA_BASE is  %08x \n", IO_DMA_BASE));
  KGDB_PRINTF(("dma_base addr is %02x %02x %02x %02x \n", iodma_base->addr_1, iodma_base->addr_2, iodma_base->addr_3, iodma_base->addr_4));
  KGDB_PRINTF(("DMA_XMT_BASE_0 is %08x \n", DMA_XMT_BASE_0));
  KGDB_PRINTF(("DMA_XMT_BASE_1 is %08x \n", DMA_XMT_BASE_1));
  KGDB_PRINTF(("DMA_RCV_BASE   is %08x \n", DMA_RCV_BASE  ));
#endif  
}


void rcv_init()
{
  edma->rcvcs = RST;
  
#define DEBUG_RCVCS 1
#if DEBUG_RCVCS
  KGDB_PRINTF(("after rst, edma->rcvcs is %02x \n", edma->rcvcs));
  KGDB_PRINTF(("hp is %08x \n", (DMA_RCV_BASE + (edma->rcvhp << 8))));
  KGDB_PRINTF(("tp is %08x \n", (DMA_RCV_BASE + (edma->rcvtp << 8))));  
#endif

  if (edma->rcvcs & OVERRUN) {
	  KGDB_PRINTF(("%s detects overrun \n",__FUNCTION__));
  } else {
	  edma->rcvcs = DMARUN | IE; /*  ???????????? */
  }
#if DEBUG_RCVCS
  KGDB_PRINTF(("after run, edma->rcvcs is %02x \n", edma->rcvcs));
#endif
}


void xmt_init()
{
  edma->xmtcs = RST;
#define DEBUG_XMTCS 1
#if DEBUG_XMTCS
  KGDB_PRINTF(("after rst, edma->xmtcs is %02x \n", edma->xmtcs));
#endif

  edma->xmtcs = DMARUN;
#if DEBUG_XMTCS
  KGDB_PRINTF(("after run, edma->xmtcs is %02x \n", edma->xmtcs));
#endif
  
}

#if DEBUG
void printbuf(buf, len)
unsigned char *buf;
int len;
{
  int i, j = 0;

  if (landebug & LANDEBUG_VERBOSE) {
	  kgdb_printf("\n");
	  for (i = 0; i < len; i++) {
		  kgdb_printf("%02x ",buf[i]);
		  j++;
		  if (j%16 == 0) {
			  kgdb_printf("\n");
			  j = 0;
		  }
	  }
	  kgdb_printf("\n");
	  kgdb_printf("total bytes is %04x\n", len);
  }
}
#endif /* DEBUG */

