/*
 * Copyright 1996 1995 by Open Software Foundation, Inc. 1997 1996 1995 1994 1993 1992 1991  
 *              All Rights Reserved 
 *  
 * Permission to use, copy, modify, and distribute this software and 
 * its documentation for any purpose and without fee is hereby granted, 
 * provided that the above copyright notice appears in all copies and 
 * that both the copyright notice and this permission notice appear in 
 * supporting documentation. 
 *  
 * OSF DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE 
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
 * FOR A PARTICULAR PURPOSE. 
 *  
 * IN NO EVENT SHALL OSF BE LIABLE FOR ANY SPECIAL, INDIRECT, OR 
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM 
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, 
 * NEGLIGENCE, OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION 
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 
 */
/*
 * Copyright 1996 1995 by Apple Computer, Inc. 1997 1996 1995 1994 1993 1992 1991  
 *              All Rights Reserved 
 *  
 * Permission to use, copy, modify, and distribute this software and 
 * its documentation for any purpose and without fee is hereby granted, 
 * provided that the above copyright notice appears in all copies and 
 * that both the copyright notice and this permission notice appear in 
 * supporting documentation. 
 *  
 * APPLE COMPUTER DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE 
 * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS 
 * FOR A PARTICULAR PURPOSE. 
 *  
 * IN NO EVENT SHALL APPLE COMPUTER BE LIABLE FOR ANY SPECIAL, INDIRECT, OR 
 * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM 
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN ACTION OF CONTRACT, 
 * NEGLIGENCE, OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION 
 * WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 
 */
/*
 * MkLinux_DR2.1_Update1
 */

#include <lan.h>
#include <mach_kgdb.h>
#include <debug.h>
#include <mach_debug.h>
#include <kern/misc_protos.h>
#include <kern/assert.h>

#include <kgdb/gdb_defs.h>
#include <kgdb/kgdb_defs.h>	/* For kgdb_printf */

#include <ppc/spl.h>
#include <ppc/proc_reg.h>
#include <ppc/misc_protos.h>
#include <ppc/trap.h>
#include <ppc/exception.h>
#include <ppc/POWERMAC/serial_io.h>
#include <ppc/POWERMAC/rtclock_entries.h>
#include <ppc/POWERMAC/powermac.h>

/* NMGS TODO this SPL stuff needs thinking out */

spl_t sploff(void)	{ return set_priority_level(SPLOFF); }
spl_t splhigh(void)	{ return set_priority_level(SPLHIGH); }
spl_t splsched(void)	{ return set_priority_level(SPLSCHED); }
spl_t splclock(void)	{ return set_priority_level(SPLCLOCK); }
spl_t splpower(void)	{ return set_priority_level(SPLPOWER); }
spl_t splvm(void)	{ return set_priority_level(SPLVM); }
spl_t splbio(void)	{ return set_priority_level(SPLBIO); }
spl_t splimp(void)	{ return set_priority_level(SPLIMP); }
spl_t spltty(void)	{ return set_priority_level(SPLTTY); }
spl_t splnet(void)	{ return set_priority_level(SPLNET); }
spl_t splsclk(void)	{ return set_priority_level(SPLSCLK); }

void  spllo(void)	{ (void) set_priority_level(SPLLO); }
void  splx(spl_t level) { (void) set_priority_level(level); }
void  splon(spl_t level){ (void) set_priority_level(level); }

spl_t current_priority = SPLHIGH;	/* MEB 11/2/95  */

#if DEBUG
vm_offset_t spl_addr;
#endif /* MACH_DEBUG */

void	(*platform_interrupt)(int type, struct ppc_saved_state *ssp,
			unsigned int dsisr, unsigned int dar);

void	(*platform_interrupt_initialize)(void);
void	(*pmac_register_int)(int device, void (*handler)(int, void *));

/* 
 * set_priority_level implements spl()
 * The lower the number, the higher the priority. TODO NMGS change?? 
 */

#if MACH_DEBUG
#if DEBUG
struct {
	spl_t level;
	char* name;
} spl_names[] = {
	{ SPLHIGH,	"SPLHIGH" },
	{ SPLCLOCK,	"SPLCLOCK" },
	{ SPLPOWER,     "SPLPOWER" },
	{ SPLVM,	"SPLVM" },
	{ SPLBIO,	"SPLBIO" },
	{ SPLIMP,	"SPLIMP" },
	{ SPLTTY,	"SPLTTY" },
	{ SPLNET,	"SPLNET" },
	{ SPLSCLK,	"SPLSCLK" },
	{ SPLLO,	"SPLLO" },
};

#define SPL_NAMES_COUNT (sizeof (spl_names)/sizeof(spl_names[0]))

static char *spl_name(spl_t lvl);

/* A routine which, given the number of an spl, returns its name.
 * this uses a table set up in spl.h in order to make the association
 */
static char *spl_name(spl_t lvl)
{
	int i;
	for (i = 0; i < SPL_NAMES_COUNT; i++)
		if (spl_names[i].level == lvl)
			return spl_names[i].name;
	printf("UNKNOWN SPL %d",lvl);
	panic("");
	return ("UNKNOWN SPL");
}
#endif /* DEBUG */

#endif /* MACH_DEBUG */


/* The implementation of splxxx() */

spl_t set_priority_level(spl_t lvl)
{
	spl_t old_level;

/*	DPRINTF(("changing priority %s from %s to %s\n",
		 (lvl > current_priority) ? "DOWN" : "UP",
		 spl_name(current_priority),
		 spl_name(lvl)));
*/
	old_level = current_priority;

	/* MEB - 11/2/95 Interrupt Hacks */

	if (lvl == SPLLO) {
		current_priority = lvl;
		interrupt_enable();
	} else {
		interrupt_disable();
		current_priority = lvl;
	}

	/* Make sure that the write completes before we continue */

	sync();

	return old_level;
}

void
interrupt_init(void)
{
	interrupt_disable();

	switch (powermac_info.class) {
	case POWERMAC_CLASS_PDM:
		platform_interrupt_initialize = pdm_interrupt_initialize;
		break;
	case POWERMAC_CLASS_PCI:
		platform_interrupt_initialize = pci_interrupt_initialize;
		break;
	default:
		panic("Unsupported class for interrupt dispatcher\n");
	}		
	(*platform_interrupt_initialize)();
	current_priority = SPLLO;
	interrupt_enable();
}


struct ppc_saved_state * interrupt(
        int type,
        struct ppc_saved_state *ssp,
	unsigned int dsisr,
	unsigned int dar)
{
        unsigned int sp, dec;
	int irq;
	spl_t		old_spl = current_priority;

	current_priority = SPLHIGH;

#if DEBUG
	{
		/* make sure we're not near to overflowing intstack */
		unsigned int sp;
		extern unsigned int intstack;
		static unsigned int spmin = 0xFFFFFFFF;
		__asm__ volatile("mr	%0, 1" : "=r" (sp));
		if (sp < (intstack + PPC_PGBYTES)) {
			printf("INTERRUPT - LOW ON STACK!\n");
			call_kgdb_with_ctx(type, 0, ssp);
		}
#if 0
		if (sp < spmin) {
			printf("INTERRUPT - NEW LOW WATER 0x%08x\n",sp);
			spmin = sp;
		}
#endif /* 0 */
	}
#endif /* DEBUG */
#if DEBUG
	if (old_spl != SPLLO) {
		kgdb_printf("{Raised SPL=%s, iptr = 0x%08x, spl was from 0x%08x??}", spl_name(old_spl), ssp->srr0, spl_addr);
	}
#endif
	switch (type) {
	case EXC_DECREMENTER:
#if	!RTC_USES_VIA
		rtclock_intr(0,ssp);  /* Hardcoded on decrementer interrupt */
#endif	/* !RTC_USES_VIA */
		goto out;

	case EXC_INTERRUPT:
		break; /* Treat below */
	default:
#if MACH_KGDB
		kgdb_printf("Received illegal interrupt type 0x%x, dar=0x%x, dsisr=0x%x old_spl=%d\n",type,dar,dsisr,old_spl);
#if DEBUG
		regDump(ssp);
#endif /* DEBUG */
		call_kgdb_with_ctx(type, 0, ssp);
		goto out;
#else
		/* if we haven't got a debugger,
		 * try ignoring illegal interrupt
		 */
		goto out;
#endif /* MACH_KGDB */
	}

	/* Call the platform interrupt routine */
	platform_interrupt(type, ssp, dsisr, dar);

out:
	current_priority = old_spl;
	return ssp;
}
